@@ -22,32 +22,11 @@
 #include "gconf-sources.h"
 #include "gconf-locale.h"
 #include "gconfd.h"
+#include <string.h>
 #include <unistd.h>
 #include <time.h>
 #include <stdlib.h>
 
-/* This makes hash table safer when debugging */
-#ifndef GCONF_ENABLE_DEBUG
-#define safe_g_hash_table_insert g_hash_table_insert
-#else
-static void
-safe_g_hash_table_insert(GHashTable* ht, gpointer key, gpointer value)
-{
-  gpointer oldkey = NULL, oldval = NULL;
-
-  if (g_hash_table_lookup_extended(ht, key, &oldkey, &oldval))
-    {
-      gconf_log(GCL_WARNING, "Hash key `%s' is already in the table!",
-                (gchar*) key);
-      return;
-    }
-  else
-    {
-      g_hash_table_insert(ht, key, value);
-    }
-}
-#endif
-
 /*
  * Forward decls
  */
@@ -113,7 +92,7 @@ impl_ConfigDatabase_lookup_with_locale(PortableServer_Servant servant,
   gboolean is_writable = TRUE;
 
   if (gconfd_check_in_shutdown (ev))
-    return invalid_corba_value ();
+    return gconf_invalid_corba_value ();
   
   locale_list = locale_cache_lookup(locale);
   
@@ -131,7 +110,7 @@ impl_ConfigDatabase_lookup_with_locale(PortableServer_Servant servant,
   
   if (val != NULL)
     {
-      ConfigValue* cval = corba_value_from_gconf_value(val);
+      ConfigValue* cval = gconf_corba_value_from_gconf_value(val);
 
       gconf_value_free(val);
 
@@ -143,7 +122,7 @@ impl_ConfigDatabase_lookup_with_locale(PortableServer_Servant servant,
     {
       gconf_set_exception(&error, ev);
 
-      return invalid_corba_value();
+      return gconf_invalid_corba_value ();
     }
 }
 
@@ -153,7 +132,7 @@ impl_ConfigDatabase_lookup(PortableServer_Servant servant,
                            CORBA_Environment * ev)
 {
   if (gconfd_check_in_shutdown (ev))
-    return invalid_corba_value ();
+    return gconf_invalid_corba_value ();
   
   return impl_ConfigDatabase_lookup_with_locale (servant, key,
                                                  NULL, TRUE, NULL,
@@ -172,7 +151,7 @@ impl_ConfigDatabase_lookup_default_value(PortableServer_Servant servant,
   GConfLocaleList* locale_list;  
 
   if (gconfd_check_in_shutdown (ev))
-    return invalid_corba_value ();
+    return gconf_invalid_corba_value ();
   
   locale_list = locale_cache_lookup(locale);
   
@@ -185,7 +164,7 @@ impl_ConfigDatabase_lookup_default_value(PortableServer_Servant servant,
   
   if (val != NULL)
     {
-      ConfigValue* cval = corba_value_from_gconf_value(val);
+      ConfigValue* cval = gconf_corba_value_from_gconf_value(val);
 
       gconf_value_free(val);
 
@@ -197,7 +176,7 @@ impl_ConfigDatabase_lookup_default_value(PortableServer_Servant servant,
     {
       gconf_set_exception(&error, ev);
 
-      return invalid_corba_value();
+      return gconf_invalid_corba_value ();
     }
 }
 
@@ -221,7 +200,6 @@ impl_ConfigDatabase_set(PortableServer_Servant servant,
                         CORBA_Environment * ev)
 {
   GConfDatabase *db = (GConfDatabase*) servant;
-  gchar* str;
   GConfValue* val;
   GError* error = NULL;
 
@@ -242,15 +220,16 @@ impl_ConfigDatabase_set(PortableServer_Servant servant,
       return;
     }
       
-  str = gconf_value_to_string(val);
-
 #if 0
-  /* reduce traffice to the logfile */
-  gconf_log(GCL_DEBUG, "Received request to set key `%s' to `%s'", key, str);
-#endif
-  
-  g_free(str);
+  {
+    gchar* str = gconf_value_to_string(val);
 
+    /* reduce traffice to the logfile */
+    gconf_log(GCL_DEBUG, "Received request to set key `%s' to `%s'", key, str);
+
+    g_free(str);
+  }
+#endif
   
   gconf_database_set(db, key, val, value, &error);
 
@@ -260,10 +239,10 @@ impl_ConfigDatabase_set(PortableServer_Servant servant,
 }
 
 static void
-impl_ConfigDatabase_unset_with_locale(PortableServer_Servant servant,
-				      const CORBA_char * key,
-				      const CORBA_char * locale,
-				      CORBA_Environment * ev)
+impl_ConfigDatabase_unset_with_locale (PortableServer_Servant servant,
+                                       const CORBA_char * key,
+                                       const CORBA_char * locale,
+                                       CORBA_Environment * ev)
 {
   GConfDatabase *db = (GConfDatabase*) servant;
   GError* error = NULL;
@@ -405,7 +384,8 @@ impl_ConfigDatabase_all_entries(PortableServer_Servant servant,
       g_assert(p->key != NULL);
 
       (*keys)->_buffer[i] = CORBA_string_dup(p->key);
-      fill_corba_value_from_gconf_value(p->value, &((*values)->_buffer[i]));
+      gconf_fill_corba_value_from_gconf_value (gconf_entry_get_value (p),
+                                               &((*values)->_buffer[i]));
       (*is_defaults)->_buffer[i] = gconf_entry_get_is_default(p);
       (*is_writables)->_buffer[i] = gconf_entry_get_is_writable(p);
       
@@ -475,10 +455,10 @@ impl_ConfigDatabase_all_dirs(PortableServer_Servant servant,
 }
 
 static void
-impl_ConfigDatabase_set_schema(PortableServer_Servant servant,
-			       const CORBA_char * key,
-			       const CORBA_char * schema_key,
-			       CORBA_Environment * ev)
+impl_ConfigDatabase_set_schema (PortableServer_Servant servant,
+                                const CORBA_char * key,
+                                const CORBA_char * schema_key,
+                                CORBA_Environment * ev)
 {
   GConfDatabase *db = (GConfDatabase*) servant;
   GError* error = NULL;
@@ -486,14 +466,17 @@ impl_ConfigDatabase_set_schema(PortableServer_Servant servant,
   if (gconfd_check_in_shutdown (ev))
     return;
   
-  gconf_database_set_schema(db, key, schema_key, &error);
-
-  gconf_set_exception(&error, ev);
+  gconf_database_set_schema (db, key,
+                             *schema_key != '\0' ?
+                             schema_key : NULL,
+                             &error);
+  
+  gconf_set_exception (&error, ev);
 }
 
 static void
-impl_ConfigDatabase_sync(PortableServer_Servant servant,
-			 CORBA_Environment * ev)
+impl_ConfigDatabase_sync (PortableServer_Servant servant,
+                          CORBA_Environment * ev)
 {
   GConfDatabase *db = (GConfDatabase*) servant;
   GError* error = NULL;
@@ -501,9 +484,9 @@ impl_ConfigDatabase_sync(PortableServer_Servant servant,
   if (gconfd_check_in_shutdown (ev))
     return;
   
-  gconf_database_sync(db, &error);
+  gconf_database_sync (db, &error);
 
-  gconf_set_exception(&error, ev);
+  gconf_set_exception (&error, ev);
 }
 
 static void
@@ -516,16 +499,16 @@ impl_ConfigDatabase_clear_cache(PortableServer_Servant servant,
   if (gconfd_check_in_shutdown (ev))
     return;
   
-  gconf_log(GCL_DEBUG, _("Received request to drop all cached data"));  
+  gconf_log (GCL_DEBUG, _("Received request to drop all cached data"));  
   
-  gconf_database_clear_cache(db, &error);
+  gconf_database_clear_cache (db, &error);
 
-  gconf_set_exception(&error, ev);
+  gconf_set_exception (&error, ev);
 }
 
 static void
-impl_ConfigDatabase_synchronous_sync(PortableServer_Servant servant,
-				     CORBA_Environment * ev)
+impl_ConfigDatabase_synchronous_sync (PortableServer_Servant servant,
+                                      CORBA_Environment * ev)
 {
   GConfDatabase *db = (GConfDatabase*) servant;
   GError* error = NULL;
@@ -536,9 +519,9 @@ impl_ConfigDatabase_synchronous_sync(PortableServer_Servant servant,
   gconf_log(GCL_DEBUG, _("Received request to sync synchronously"));
   
   
-  gconf_database_synchronous_sync(db, &error);
+  gconf_database_synchronous_sync (db, &error);
 
-  gconf_set_exception(&error, ev);
+  gconf_set_exception (&error, ev);
 }
 
 
@@ -562,7 +545,7 @@ impl_ConfigDatabase2_lookup_with_schema_name(PortableServer_Servant servant,
   ConfigValue* cval;
   
   if (gconfd_check_in_shutdown (ev))
-    return invalid_corba_value ();
+    return gconf_invalid_corba_value ();
   
   locale_list = locale_cache_lookup(locale);
 
@@ -588,13 +571,13 @@ impl_ConfigDatabase2_lookup_with_schema_name(PortableServer_Servant servant,
   
   if (val != NULL)
     {
-      cval = corba_value_from_gconf_value(val);
+      cval = gconf_corba_value_from_gconf_value(val);
       gconf_value_free(val);
       g_return_val_if_fail(error == NULL, cval);
     }
   else
     {
-      cval = invalid_corba_value();
+      cval = gconf_invalid_corba_value ();
     }
 
   gconf_log (GCL_DEBUG, "In lookup_with_schema_name returning schema name '%s' error '%s'",
@@ -684,8 +667,9 @@ impl_ConfigDatabase2_all_entries_with_schema_name(PortableServer_Servant servant
       g_assert(p != NULL);
       g_assert(p->key != NULL);
 
-      (*keys)->_buffer[i] = CORBA_string_dup(p->key);
-      fill_corba_value_from_gconf_value(p->value, &((*values)->_buffer[i]));
+      (*keys)->_buffer[i] = CORBA_string_dup (p->key);
+      gconf_fill_corba_value_from_gconf_value (gconf_entry_get_value (p),
+                                               &((*values)->_buffer[i]));
       (*schema_names)->_buffer[i] = CORBA_string_dup (gconf_entry_get_schema_name (p));
       if ((*schema_names)->_buffer[i] == NULL)
         (*schema_names)->_buffer[i] = CORBA_string_dup ("");
@@ -711,11 +695,11 @@ impl_ConfigDatabase3_add_listener_with_properties (PortableServer_Servant servan
                                                    CORBA_Environment *ev)
 {
   GConfDatabase *db = (GConfDatabase*) servant;
-  const char *name;
+  const char *name = NULL;
   int i;
   
   if (gconfd_check_in_shutdown (ev))
-    return;
+    return 0;
 
   i = 0;
   while (i < (int) properties->_length)
@@ -732,12 +716,24 @@ impl_ConfigDatabase3_add_listener_with_properties (PortableServer_Servant servan
 static void
 impl_ConfigDatabase3_recursive_unset (PortableServer_Servant servant,
                                       const CORBA_char *key,
+                                      ConfigDatabase3_UnsetFlags flags,
                                       CORBA_Environment *ev)
 {
+  GConfDatabase *db = (GConfDatabase*) servant;
+  GError *error;
+  GConfUnsetFlags gconf_flags;
+  
   if (gconfd_check_in_shutdown (ev))
     return;
 
-  /* FIXME not implemented yet */
+  gconf_flags = 0;
+  if (flags & ConfigDatabase3_UNSET_INCLUDING_SCHEMA_NAMES)
+    gconf_flags |= GCONF_UNSET_INCLUDING_SCHEMA_NAMES;
+  
+  error = NULL;
+  gconf_database_recursive_unset (db, key, NULL, gconf_flags, &error);
+
+  gconf_set_exception (&error, ev);
 }
 
 static PortableServer_ServantBase__epv base_epv = {
@@ -1058,7 +1054,12 @@ gconf_database_add_listener    (GConfDatabase       *db,
        * Because it's likely the right thing for the client
        * app to simply continue.
        */
-      gconf_log (GCL_WARNING, _("Failed to log addition of listener %s (%s); will not be able to restore this listener on gconfd restart, resulting in unreliable notification of configuration changes."), err->message);
+      gconf_log (GCL_WARNING,
+		 _("Failed to log addition of listener %s (%s);"
+		   "will not be able to restore this listener on "
+		   "gconfd restart, resulting in unreliable "
+		   "notification of configuration changes."),
+		 name, err->message);
       g_error_free (err);
     }
   
@@ -1069,7 +1070,10 @@ void
 gconf_database_remove_listener (GConfDatabase       *db,
                                 CORBA_unsigned_long  cnxn)
 {
-  Listener *l = NULL;
+  union {
+      Listener *l;
+      gpointer l_ptr;
+  } l = { NULL };
   GError *err;
   const gchar *location = NULL;
 
@@ -1082,7 +1086,7 @@ gconf_database_remove_listener (GConfDatabase       *db,
   gconf_log(GCL_DEBUG, "Removing listener %u", (guint)cnxn);
 
   if (!gconf_listeners_get_data (db->listeners, cnxn,
-                                 (gpointer*)&l,
+                                 &(l.l_ptr),
                                  &location))
     {
       gconf_log (GCL_WARNING, _("Listener ID %lu doesn't exist"),
@@ -1092,12 +1096,12 @@ gconf_database_remove_listener (GConfDatabase       *db,
   else
     {
       gconf_log (GCL_DEBUG, "Name of listener %u is %s",
-                 (guint) cnxn, l->name);
+                 (guint) cnxn, l.l->name);
     }
   
   err = NULL;
   if (!gconfd_logfile_change_listener (db, FALSE, cnxn,
-                                       l->obj, location, &err))
+                                       l.l->obj, location, &err))
     {
       gconf_log (GCL_WARNING, _("Failed to log removal of listener to logfile (most likely harmless, may result in a notification weirdly reappearing): %s"),
                  err->message);
@@ -1322,7 +1326,12 @@ gconf_database_unset (GConfDatabase      *db,
       GConfValue* def_value;
       const gchar* locale_list[] = { NULL, NULL };
       gboolean is_writable = TRUE;
-      
+
+      /* This is a somewhat dubious optimization
+       * that assumes that if the unset was successful
+       * the default value is the new value. Which is
+       * safe for now, I _think_
+       */
       locale_list[0] = locale;
       def_value = gconf_database_query_default_value(db,
                                                      key,
@@ -1336,12 +1345,12 @@ gconf_database_unset (GConfDatabase      *db,
 
       if (def_value != NULL)
         {
-          val = corba_value_from_gconf_value(def_value);
+          val = gconf_corba_value_from_gconf_value(def_value);
           gconf_value_free(def_value);
         }
       else
         {
-          val = invalid_corba_value();
+          val = gconf_invalid_corba_value ();
         }
           
       gconf_database_schedule_sync(db);
@@ -1352,6 +1361,97 @@ gconf_database_unset (GConfDatabase      *db,
     }
 }
 
+void
+gconf_database_recursive_unset (GConfDatabase      *db,
+                                const gchar        *key,
+                                const gchar        *locale,
+                                GConfUnsetFlags     flags,
+                                GError            **err)
+{
+  ConfigValue* val;
+  GError* error = NULL;
+  GSList *notifies;
+  GSList *tmp;
+  
+  g_return_if_fail (err == NULL || *err == NULL);
+  
+  g_assert (db->listeners != NULL);
+  
+  db->last_access = time (NULL);
+  
+  gconf_log (GCL_DEBUG, "Received request to recursively unset key \"%s\"", key);
+
+  notifies = NULL;
+  gconf_sources_recursive_unset (db->sources, key, locale,
+                                 flags, &notifies, &error);
+
+  /* We return the error but go ahead and finish the unset.
+   * We're just returning the first error seen during the
+   * unset process.
+   */
+  if (error != NULL)
+    {
+      gconf_log (GCL_ERR, _("Error unsetting \"%s\": %s"),
+                 key, error->message);
+
+      if (err)
+        *err = error;
+      else
+        g_error_free (error);
+
+      error = NULL;
+    }
+  
+  tmp = notifies;
+  while (tmp != NULL)
+    {
+      GConfValue* new_value;
+      const gchar* locale_list[] = { NULL, NULL };
+      gboolean is_writable = TRUE;
+      gboolean is_default = TRUE;
+      char *notify_key = tmp->data;
+      
+
+      locale_list[0] = locale;
+      new_value = gconf_database_query_value (db,
+                                              notify_key,
+                                              locale_list,
+                                              TRUE,
+                                              NULL,
+                                              &is_default,
+                                              &is_writable,
+                                              &error);
+
+      if (error)
+        gconf_log (GCL_ERR, _("Error getting new value for \"%s\": %s"),
+                   notify_key, error->message);
+      g_propagate_error (err, error);
+      error = NULL;
+      
+      if (new_value != NULL)
+        {
+          val = gconf_corba_value_from_gconf_value (new_value);
+          gconf_value_free (new_value);
+        }
+      else
+        {
+          val = gconf_invalid_corba_value ();
+        }
+          
+      gconf_database_schedule_sync (db);
+
+      gconf_database_notify_listeners (db, notify_key, val,
+                                       is_default, is_writable);
+      
+      CORBA_free (val);
+      g_free (notify_key);
+      
+      tmp = tmp->next;
+    }
+
+  g_slist_free (notifies);
+}
+
 gboolean
 gconf_database_dir_exists  (GConfDatabase  *db,
                             const gchar    *dir,
@@ -1389,13 +1489,13 @@ gconf_database_remove_dir  (GConfDatabase  *db,
   
   db->last_access = time(NULL);
   
-  gconf_log(GCL_DEBUG, "Received request to remove dir `%s'", dir);
+  gconf_log (GCL_DEBUG, "Received request to remove directory \"%s\"", dir);
   
   gconf_sources_remove_dir(db->sources, dir, err);
 
   if (err && *err != NULL)
     {
-      gconf_log(GCL_ERR, _("Error removing dir `%s': %s"),
+      gconf_log (GCL_ERR, _("Error removing directory \"%s\": %s"),
                  dir, (*err)->message);
     }
   else
@@ -1442,13 +1542,13 @@ gconf_database_all_dirs (GConfDatabase  *db,
   
   db->last_access = time(NULL);
     
-  gconf_log(GCL_DEBUG, "Received request to list all subdirs in `%s'", dir);
+  gconf_log (GCL_DEBUG, "Received request to list all subdirs in `%s'", dir);
 
   subdirs = gconf_sources_all_dirs (db->sources, dir, err);
 
   if (err && *err != NULL)
     {
-      gconf_log(GCL_ERR, _("Error listing dirs in `%s': %s"),
+      gconf_log (GCL_ERR, _("Error listing dirs in `%s': %s"),
                  dir, (*err)->message);
     }
   return subdirs;
@@ -1458,23 +1558,23 @@ void
 gconf_database_set_schema (GConfDatabase  *db,
                            const gchar    *key,
                            const gchar    *schema_key,
-                           GError    **err)
+                           GError        **err)
 {
-  g_return_if_fail(err == NULL || *err == NULL);
-  g_assert(db->listeners != NULL);
+  g_return_if_fail (err == NULL || *err == NULL);
+  g_assert (db->listeners != NULL);
   
-  db->last_access = time(NULL);
+  db->last_access = time (NULL);
   
-  gconf_sources_set_schema(db->sources, key, schema_key, err);
+  gconf_sources_set_schema (db->sources, key, schema_key, err);
 
   if (err && *err != NULL)
     {
-      gconf_log(GCL_ERR, _("Error setting schema for `%s': %s"),
+      gconf_log (GCL_ERR, _("Error setting schema for `%s': %s"),
                 key, (*err)->message);
     }
   else
     {
-      gconf_database_schedule_sync(db);
+      gconf_database_schedule_sync (db);
     }
 }
 
